-----------------------------------------------------------------------------------------------
-- Client Lua Script for DeathReport
-- Copyright (c) Tyler T. Hardy. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- DeathReport Module Definition
-----------------------------------------------------------------------------------------------
local DeathReport = {} 
--local Comms = Apollo.GetPackage("PotatoComms").tPackage:new(DeathReport)

local bInPVP = false

local bDebug = true

local DRVariables = {
	["%player"] = {"Player that was killed.","strTargetName"},			--CUSTOM
	["%enemy"] = {"Enemy that caused the death.","strCasterName"},		--CUSTOM
	["%spell"] = {"Spell that caused the death.","strSpell"},			--CUSTOM
	["%raw"] = {"Raw damage of killing blow.","nRawDamage"},
	["%damage"] = {"Damage after armor + shield.","nDamageAmount"},
	["%shield"] = {"Damage shielded.","nShield"},
	["%absorb"] = {"Damage absorbed.","nAbsorption"},
	["%overkill"] = {"Overkill damage.","nOverkill"},
	["%dot"] = {"+ if killing blow is a dot.","bPeriodic"},
	["%crit"] = {"* if killing blow is critical.","bCritical"},
	["%aggro"] = {"[AGGRO] if player had aggro.","bAggro"}, 			--CUSTOM
}
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999

function DeathReport:IsTableEmpty(tTable)
	for key,val in pairs(tTable) do
		return false
	end
	return true
end 

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function DeathReport:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function DeathReport:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- DeathReport OnLoad
-----------------------------------------------------------------------------------------------
function DeathReport:OnLoad() --Loaded 1st
  	--load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("DeathReport.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

function DeathReport:OnDocLoaded() --Loaded 3rd
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.tDeaths = {}
		self.tGroup = {}
		
		if self.Settings == nil or self.Settings.Group == nil or self.Settings.Report == nil then
			self.Settings = {
				Group = {
					PvP = {Report=false,Recap=false},
					Party = {Report=true,Recap=false},
					Dungeon = {Report=true,Recap=false},
					Raid = {Report=true,Recap=true}
				},
				Report = "%player was killed by %enemy with %spell (%damage%crit)%aggro.",
				MaxLogs = 10
			}
		end
		
		--Windows
	    self.wndRecap = Apollo.LoadForm(self.xmlDoc, "DeathForm", nil, self)
		if self.Settings.wndRecapPos ~= nil then
			self.wndRecap:SetAnchorOffsets(unpack(self.Settings.wndRecapPos))
		end
		
		self.wndSettings = self.wndRecap:FindChild("Settings")		
		self.wndRecap:FindChild("SettingsBtn"):AttachWindow(self.wndRecap:FindChild("Settings"))
		local wndGroupSettings = self.wndSettings:FindChild("GroupSettings")
	
		for strGroupType,tGroupSettings in pairs(self.Settings.Group) do
			local wndGroupType = wndGroupSettings:FindChild(strGroupType)
		
			for strRR,bState in pairs(tGroupSettings) do
				if strGroupType ~= "PvP" then
					wndGroupType:FindChild(strRR):SetCheck(bState)
				else
					wndGroupType:FindChild("Report"):SetCheck(false)
					wndGroupType:FindChild("Report"):Enable(false)
				end
			end
		end
		
		self.wndVariables = Apollo.LoadForm(self.xmlDoc, "Variables", nil, self)
		for strShortcut,tVar in pairs(DRVariables) do
			local wndVariable = Apollo.LoadForm(self.xmlDoc, "Variable", self.wndVariables:FindChild("Container"), self)
			
			wndVariable:FindChild("Shortcut"):SetText(strShortcut)
			wndVariable:FindChild("Description"):SetText(tVar[1])
			
			self.wndVariables:FindChild("Container"):ArrangeChildrenVert()
		end
		self.wndSettings:FindChild("ReportSettings:Help"):SetTooltipForm(self.wndVariables)
		self.wndSettings:FindChild("ReportSettings:EditBox"):SetText(self.Settings.Report)
		
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterSlashCommand("deaths", "ToggleDeathWindow", self)		
		Apollo.RegisterSlashCommand("deathreport", "ToggleDeathWindow", self)		
		Apollo.RegisterSlashCommand("dr", "ToggleDeathWindow", self)		

		if GameLib.GetPlayerUnit() == nil then
			Apollo.RegisterEventHandler("CharacterCreated", "OnCharacterCreated", self)
		else
			self:OnCharacterCreated()
		end
		
		--Triggers for Deaths
		Apollo.RegisterEventHandler("CombatLogDamage", "OnDamage", self)
		Apollo.RegisterEventHandler("CombatLogFallingDamage", "OnCombatLogFallingDamage", self)
		
		--Combat for encounter tracking
		Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
	
		--PvP Stuff		
		Apollo.RegisterEventHandler("PVPMatchStateUpdated", "OnPVPMatchEntered", self)
        Apollo.RegisterEventHandler("PVPDeathmatchPoolUpdated", "OnPVPMatchEntered", self)
        Apollo.RegisterEventHandler("PVPMatchTeamInfoUpdated", "OnPVPMatchEntered", self)
        Apollo.RegisterEventHandler("PVPMatchFinished", "OnPvPMatchExited", self)
		Apollo.RegisterEventHandler("MatchExited", "OnPVPMatchExited", self)
		
		--Group stuff for determining reporter	
		Apollo.RegisterEventHandler("Group_MemberPromoted",		"OnGroupMemberPromoted", self)		-- ( name, bSelf )
		--Apollo.RegisterEventHandler("Group_Invited",			"OnGroupInvited", self)				-- ( name )
		--Apollo.RegisterEventHandler("Group_Invite_Result",		"OnGroupInviteResult", self)		-- ( name, result )
		--Apollo.RegisterEventHandler("Group_JoinRequest",		"OnGroupJoinRequest", self)			-- ( name )
		--Apollo.RegisterEventHandler("Group_Referral",			"OnGroupReferral", self)			-- ( nMemberIndex, name )
		--Apollo.RegisterEventHandler("Group_Request_Result",		"OnGroupRequestResult", self)		-- ( name, result, bIsJoin )
		Apollo.RegisterEventHandler("Group_Join",				"OnGroupJoin", self)				-- ()
		Apollo.RegisterEventHandler("Group_Left",				"OnGroupLeft", self)				-- ( reason )
		Apollo.RegisterEventHandler("Group_Add",				"OnGroupAdd", self)					-- ( name )
		Apollo.RegisterEventHandler("Group_Remove",				"OnGroupRemove", self)				-- ( name, reason )
				
		-- Do additional Addon initialization here
	end
end

function DeathReport:OnSave(eLevel)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Account then return end
	
	local tData = TableUtil:Copy(self.Settings)
	--self.Settings.wndRecapPos = {self.wndRecap:GetAnchorOffsets()}
	
	return tData
end

function DeathReport:OnRestore(eLevel, tData) --Loaded 2nd
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Account then return end
	
	if tData then
		self.Settings = tData
	end
end

-----------------------------------------------------------------------------------------------
-- DeathReport Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

function DeathReport:ToggleDeathWindow()
	if self.wndRecap:IsShown() then
		self.wndRecap:Show(false)
		self.wndRecap:FindChild("Deaths:DeathItems"):DestroyChildren()
	else
		self:ShowRecap()
	end
end

function DeathReport:IdentifyGroup()
	if bInPVP then
		return "PvP"
	elseif GroupLib.InRaid() then
		return "Raid"
	elseif GroupLib.InInstance() then
		return "Dungeon"
	elseif GroupLib.InGroup() then
		return "Party"
	else
		return "None"
	end
end

function DeathReport:ShouldReport()
	local strGroup = self:IdentifyGroup()
	if strGroup == "None" then
		return false
	else
		return self.bReporter and self.Settings.Group[strGroup].Report
	end
end

function DeathReport:ShouldRecap()
	local strGroup = self:IdentifyGroup()
	if strGroup == "None" then
		return bDebug
	else
		return self.Settings.Group[strGroup].Recap
	end
end

---------------------------------------------------------------------------------------------------
-- Event Handler Functions
---------------------------------------------------------------------------------------------------
-- Define event functions here

function DeathReport:OnCharacterCreated()
	-- Maybe the UI reloaded so be sure to check if we are in combat or a PvP match
	if GameLib.GetPlayerUnit():IsInCombat() then
		self:OnEnteredCombat(GameLib.GetPlayerUnit(), true)
	end
	if MatchingGame:IsInMatchingGame() then
		local tMatchState = MatchingGame:GetPVPMatchState()

		if tMatchState ~= nil then
			self:OnPVPMatchEntered()
		end
	end
	
	--Group determination for reporting
	self:UpdateChannel()
	table.insert(self.tGroup, GameLib.GetPlayerUnit():GetName())
	self:UpdateGroupReporter()
	self:StartJoinRequest()
	--self:SendJoinRequest()
end

function DeathReport:OnPVPMatchEntered()
	bInPVP = true
end

function DeathReport:OnPVPMatchExited()
	bInPVP = false
end

function DeathReport:OnEnteredCombat(unitTarget, bInCombat)
	if unitTarget == GameLib.GetPlayerUnit() then
		self.bInCombat = bInCombat

		if bInCombat then
			if #self.tDeaths ~= 0 then
				self.wndRecap:FindChild("DeathItems"):DestroyChildren()
				self.tDeaths = {}
			end
						
			--Recap window
			if self.wndRecap:IsShown() then
				self.wndRecap:Show(false)
			end
		end
	end
end

function DeathReport:ShowRecap()
	self.wndRecap:Show(true)
end

function DeathReport:OnCombatLogFallingDamage(tInfo)
	--tInfo.nDamageAmount
	--tInfo.unitCaster
	if tInfo.unitCaster then
		if tInfo.unitCaster:IsDead() and tInfo.unitCaster:IsInYourGroup() then
			tInfo.strCasterName = tInfo.unitCaster and tInfo.unitCaster:GetName() or "Unknown"
			local strGender = tInfo.unitCaster:GetGender() == 1 and "her" or "his"
			
			--Log Death into Fight's Table
			if not self.tDeaths[tInfo.strCasterName] then
				self.tDeaths[tInfo.strCasterName] = tInfo
				tInfo.strTargetName = tInfo.strCasterName
				tInfo.nDamageAmount = 0
				tInfo.strSpell = "Falling Damage"
				if self:ShouldReport() then
					self:ReportToChat(tInfo.strCasterName.." fell to "..strGender.." death, like an idiot.")
				end
			end		
		end
	end
end

local tFailchecks = {
	["Seismic Tremor"]="a", --SC Thunderfoot
}

function DeathReport:OnDamage(tInfo)
	if tInfo.unitCaster and tInfo.unitTarget then
		tInfo.strSpell = tInfo.splCallingSpell and tInfo.splCallingSpell:GetName() or "Unknown"
		if tInfo.bTargetKilled and tInfo.unitTarget:IsInYourGroup() then
			--Setup Custom Variables
			tInfo.strTargetName = tInfo.unitTarget and tInfo.unitTarget:GetName() or "Unknown"
			tInfo.strCasterName = tInfo.unitCaster and tInfo.unitCaster:GetName() or "Unknown"	
			tInfo.bCritical = tInfo.eCombatResult == GameLib.CodeEnumCombatResult.Critical
			if tInfo.unitCaster then
				tInfo.bAggro = tInfo.unitCaster:GetTarget() == tInfo.unitTarget
			end
			
			--Setup Player Info if Needed (Class, HP, etc.)
			tInfo.nClass = tInfo.unitTarget:GetClassId()
			
			--Clear Memory Intensive Variables (TODO: Test without removal.)
			tInfo.unitCaster = nil
			tInfo.unitTarget = nil
			tInfo.splCallingSpell = nil
			
			--Log Death into Fight's Table
			self:RegisterDeath(tInfo)
						
			--Report death
			if self:ShouldReport() then
				self:ReportToChat(self:GetReportString(tInfo))
			end
		end
		
		--[[Check for failchecks TODO: Remove
		if tFailchecks[tInfo.strSpell] then
			local strPlayerName = tInfo.unitTarget:GetName() or tInfo.strTargetName
			self:ReportToChat(strPlayerName.." hit by " .. tInfo.strSpell..". FUCKIN FAIL!")
		end]]--
	end
end

function DeathReport:RegisterDeath(tInfo)
	table.insert(self.tDeaths, tInfo)
	
	local wndDeaths = self.wndRecap:FindChild("DeathItems")
	local wndDeath = Apollo.LoadForm(self.xmlDoc, "DeathItem", wndDeaths, self)
	
	wndDeath:FindChild("Class"):SetSprite("IconSprites:Icon_Windows_UI_CRB_"..self:GetClassById(tInfo.nClass))
	wndDeath:FindChild("Name"):SetText(tInfo.strTargetName)
	wndDeath:FindChild("Source"):SetText(tInfo.strCasterName)
	wndDeath:FindChild("Attack"):SetText(tInfo.strSpell)
	wndDeath:FindChild("Damage"):SetText(self:GetFormattedVariable(tInfo, "nDamageAmount"))
	
	wndDeaths:ArrangeChildrenVert()
end

function DeathReport:GetFormattedVariable(tInfo, strVariable)
	local vVar = tInfo[strVariable]	
	if vVar == nil then return end
	
	local strType = type(vVar)
	
	if strType == "number" then
		return string.format("%.1fk",vVar/1000)
	elseif strType == "boolean" then
		if strVariable == "bCritical" then
			return tInfo[strVariable] and "*" or ""
		elseif strVariable == "bPeriodic" then
			return tInfo[strVariable] and "+" or ""
		elseif strVariable == "bAggro" then
			return tInfo[strVariable] and " [AGGRO]" or ""
		end
	elseif strType == "function" then
		Print("[DeathReport] err Function type found for Report string.")
	else
		--String or nil (retyped as empty string)
		return vVar
	end
	
	return nil
end
	
function DeathReport:GetReportString(tInfo)	
	local strReport = self.Settings.Report
	
	for key,val in pairs(DRVariables) do
		local strFormattedVariable = self:GetFormattedVariable(tInfo, val[2])
		
		if strFormattedVariable then
			strReport = string.gsub(strReport,"%"..key, self:GetFormattedVariable(tInfo, val[2]))				
		end
	end

	return strReport
end

---------------------------------------------------------------------------------------------------
-- DeathItem Functions
---------------------------------------------------------------------------------------------------

function DeathReport:ReportDeath( wndHandler, wndControl, eMouseButton )
	local tInfo = wndHandler:GetData()
	--[[local bTargetKilled = tInfo.bTargetKilled
	local eCombatResult = tInfo.eCombatResult -- work out
	local eDamageType = tInfo.eDamageType
	--local eEffectType = tInfo.eEffectType
	local nAbsorption = tInfo.nAbsorption
	local nDamageAmount = tInfo.nDamageAmount
	local nOverkill = tInfo.nOverkill
	local nRawDamage = tInfo.nRawDamage
	local nShield = tInfo.nShield
	local splDeathSpell = tInfo.splCallingSpell
	local unitPlayer = tInfo.unitTarget
	local unitEnemy = tInfo.unitCaster]]--

	local strPlayerName = wndHandler:FindChild("Name"):GetText()
	local strEnemyName = wndHandler:FindChild("Source"):GetText()
	local strAbility = wndHandler:FindChild("Attack"):GetText()
	local strRawDamage = wndHandler:FindChild("Damage"):GetText()
	
	local strDeath = string.format("%s was killed by %s with %s (%s).", strPlayerName, strEnemyName, strAbility, strRawDamage)
	
	self:ReportToChat(strDeath)
end

---------------------------------------------------------------------------------------------------
-- DeathForm Functions
---------------------------------------------------------------------------------------------------

function DeathReport:SortByClass( wndHandler, wndControl, eMouseButton )
end

function DeathReport:SortByName( wndHandler, wndControl, eMouseButton )
end

function DeathReport:SortBySource( wndHandler, wndControl, eMouseButton )
end

function DeathReport:SortByAttack( wndHandler, wndControl, eMouseButton )
end

function DeathReport:SortByDamage( wndHandler, wndControl, eMouseButton )
end

function DeathReport:OnToggleRR( wndHandler, wndControl, eMouseButton )
	local strGroup = wndHandler:GetParent():GetName()
	local strRR = wndHandler:GetName()
	
	self.Settings.Group[strGroup][strRR] = wndHandler:IsChecked()
end

function DeathReport:CloseWnd( wndHandler, wndControl, eMouseButton )
	wndHandler:GetParent():Close()
end

function DeathReport:ChangeReportString( wndHandler, wndControl, strText )
	self.Settings.Report = strText
end

---------------------------------------------------------------------------------------------------
-- Utility Functions
---------------------------------------------------------------------------------------------------

function DeathReport:ReportToChat(strDeath)
	if GroupLib.InInstance() then
		ChatSystemLib.Command("/i "..strDeath)
	elseif GroupLib.InGroup() or GroupLib.InRaid() then
		ChatSystemLib.Command("/p "..strDeath)
	end
end

function DeathReport:GetClassById(nId)
	for key, val in pairs(GameLib.CodeEnumClass) do
		if val == nId then
			return key
		end
	end
end

-----------------------------------------------------------------------------------------------
-- DeathReport Instance
-----------------------------------------------------------------------------------------------
local DeathReportInst = DeathReport:new()
DeathReportInst:Init()

